// Betriebssysteme1 - Matthias Jauernig, 2005
// Empfnger-Programm der 12. bung
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <stdbool.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <signal.h>

bool ende = false;

void sigintfkt(int signr){
	printf("Bereite Empfaenger-Ende vor...\n");
	ende = true;
}

int main(void){
	int pipesOffen = 0;
	int rc, i, fd, flags;
	FILE *pipe[5];
	char eab[100], pipename[10];
	bool exists[5], pipeOffen[5];
	struct stat buf;
	
	// -1.) Initialwerte fr Arrays festlegen
	for(i=0; i<5; i++){
		exists[i] = false;
		pipeOffen[i] = false;
		pipe[i] = NULL;
	}
	
	// 0.) Signalfunktion fr SIGINT (Ctrl-C) bestimmen
	signal(SIGINT, sigintfkt);
	
	printf("Warte auf Sender...\n");
	// 1.) prfe stndig die relevanten Pipenamen ab und lese von ihnen
	while(!ende || pipesOffen>0){
		// 1.1) auf noch nicht geffnete Pipes prfen und diese ffnen
		for(i=0; i<5; i++){
			// 1.1.1) generiere Pipenamen
			sprintf(pipename, "p%d", i);
			// 1.1.2) prfe, ob Pipe existiert
			rc = stat(pipename, &buf);
			if(rc < 0){ // Datei existiert nicht oder Fehler
				if(errno != ENOENT){ // Fehler bei stat
					perror("stat");
				}
			}else{ // Datei existiert bereits
				if(!S_ISFIFO(buf.st_mode) && !exists[i]){ // Datei ist kein FIFO
					printf("File %s existiert, ist aber keine Pipe!\n", pipename);
					exists[i] = true;
				}else{
					if(!pipeOffen[i]){ // Datei ist FIFO und noch nicht geffnet
						pipe[i] = fopen(pipename, "r");
						if(pipe[i] == NULL){
							perror("fopen");
						}else{
							// bei ende Pipe beseitigen
							if(ende){
								if(fclose(pipe[i]) < 0){
									perror("fclose");
								}
								rc = unlink(pipename);
								if(rc < 0){
									perror("unlink");
								}
							}else{
								pipesOffen++;
								pipeOffen[i] = true;
								// 1.1.4) Nicht-blockierendes Read aktivieren
								printf("Sender '%s' hat sich angemeldet (%d"
								  " insgesamt)...\n", pipename, pipesOffen);
								fd = fileno(pipe[i]);
								flags = fcntl(fd, F_GETFL);
								if(flags < 0){
									perror("fcntl");
								}
								flags |= O_NONBLOCK;
								rc = fcntl(fd, F_SETFL, flags);
								if(rc < 0){
									perror("fcntl");
								}
							}
						}
					}else{ // 1.1.5) Pipe i geffnet, also von ihr lesen
						fd = fileno(pipe[i]);
						rc = read(fd, eab, sizeof(eab));
						if(rc==0){ // Pipe geschlossen
							pipeOffen[i] = false;
							printf("Sender '%s' hat sich abgemeldet "
							  "(%d brig)...\n", pipename, pipesOffen-1);
							if(fclose(pipe[i]) < 0){
								perror("fclose");
							}else{ // Pipe beseitigen
								pipesOffen--;
								exists[i] = false;
								rc = unlink(pipename);
								if(rc < 0){
									perror("unlink");
								}
							}
						}else if(rc > 0){ // Daten empfangen, also ausgeben
							printf("Pipe %d: '%s'\n",i,eab);
						}
					}
				}
			}
		}
	}
	printf("Keine Pipe mehr offen, mache Schluss...\n");
	return 0;
}
