// Betriebssysteme1, Matthias Jauernig, 2005
// Sender-Programm der 12. bung
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <stdbool.h>
#include <string.h>
#include <sys/stat.h>
#include <signal.h>

//#define FFLUSH() while(getchar()!='\n');

void sigpipefkt(int signr){
	printf("Pipe vom Empfaenger geschlossen, verlasse Sender...\n");
	exit(1);
}

int main(void){
	bool pipeErstellt = false, ende = false;
	int rc, i;
	char pipename[10], eab[100];
	FILE *pipe;
	struct stat buf;
	mode_t mode = 0644;
	
	// 0.) Signalfunktion fr SIGPIPE bestimmen
	signal(SIGPIPE, sigpipefkt);
	
	// 1.) erstelle Pipe
	for(i=0; i<5 && !pipeErstellt; i++){
		// 1.1) generiere Pipenamen
		sprintf(pipename, "p%d", i);
		// 1.2) prfe, ob File existiert
		rc = stat(pipename, &buf);
		if(rc < 0){ // Datei existiert nicht oder Fehler
			if(errno != ENOENT){ // Fehler bei stat
				perror("stat");
				exit(1);
			}
			// 1.3) Pipe anlegen
			rc = mkfifo(pipename, mode);
			if(rc<0){ // Fehler bei mkfifo
				perror("mkfifo");
				exit(2);
			}
			printf("Pipe mit Namen '%s' erstellt...\n", pipename);
			pipeErstellt = true;
		}else{ // Datei existiert bereits
			if(!S_ISFIFO(buf.st_mode)){
				printf("File %s existiert, ist aber keine Pipe!\n", pipename);
			}
		}
	}
	
	if(!pipeErstellt){
		printf("Zuviele Pipes geffnet, verlasse Programm...\n");
		exit(3);
	}
	
	// 2.) Pipe zum Schreiben ffnen + warten, bis Gegenseite Pipe zum Lesen ffnet
	printf("ffne Pipe zum Schreiben (warte auf Gegenseite)... ");
	fflush(stdout);
	pipe = fopen(pipename, "w");
	if(pipe == NULL){
		perror("fopen");
		exit(4);
	}
	sleep(1);
	printf("Pipe geffnet.\n");
	
	// 3.) Daten von Benutzer lesen und auf Pipe schreiben, bis EOF
	while(!ende){
		// wenn Datei durch Empfnger beseitigt, dann Ende
		if(stat(pipename, &buf) < 0){
			if(errno != ENOENT){ // Fehler bei stat
				perror("stat");
			}
			printf("Empfaenger hat Pipe geschlossen - verlasse Sender...\n");
			exit(0);
		}
		fflush(stdin);
		printf("Dateneingabe (Ende mit Ctrl-D): ");
		if(fgets(eab, 100, stdin)){ // 3.1) lesen bis Fehler oder EOF
			eab[strlen(eab)-1] = '\0';
			// 3.2) auf Pipe schreiben
			rc = (int)fwrite(eab, sizeof(eab), 1, pipe);
			if(rc == 0){ // schreiben fehlgeschlagen
				printf("fwrite fehl geschlagen...\n");
				break;
			}
			fflush(pipe);
		}else ende=true;
	}
	printf("\n\n");
	
	// 4.) Pipe schlieen, aber Datei belassen -> wird vom Empfnger beseitigt
	rc = fclose(pipe);
	if(rc == EOF){
		perror("fclose");
		exit(5);
	}
	
	return 0;
}
